{******************************************************************}
{ part of DSPack 1.05                                                     }
{                                                                  }
{ home page : http://www.progdigy.com                              }
{ email     : hgourvest@progdigy.com                               }
{                                                                  }
{ date      : 20-11-2001                                           }
{                                                                  }
{ The contents of this file are used with permission, subject to   }
{ the Mozilla Public License Version 1.1 (the "License"); you may  }
{ not use this file except in compliance with the License. You may }
{ obtain a copy of the License at                                  }
{ http://www.mozilla.org/MPL/MPL-1.1.html                          }
{                                                                  }
{ Software distributed under the License is distributed on an      }
{ "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   }
{ implied. See the License for the specific language governing     }
{ rights and limitations under the License.                        }
{                                                                  }
{ *****************************************************************}

{$WARN SYMBOL_DEPRECATED OFF}
unit DSFilters;

interface

uses Windows, Classes, Messages, Forms, SysUtils, ActiveX, Graphics, DirectShow,
 DSPack, DSUtil, WMF, WMFUtil;

const
  WM_CAPTURE_BITMAP = WM_APP + 2;

function GetBitmapSize(const pHeader: TBITMAPINFOHEADER): DWORD;
function PALETTISED(pbmi: TVIDEOINFO): boolean;
function PALETTE_ENTRIES(pbmi: TVIDEOINFO): DWORD;


//******************************************************************************
//
//  TDSPinList declaration
//  description: Enumerate all pins on a Filter
//
//******************************************************************************
type

  TDSPinList = class(TInterfaceList)
  private
    Filter: IBaseFilter;
    function  GetPin(Index: Integer): IPin;
    procedure PutPin(Index: Integer; Item: IPin);
    function  GetPinInfo(index: integer): TPin_Info;
  public
    constructor Create(BaseFilter: IBaseFilter); overload;
    destructor Destroy; override;
    procedure Update;
    procedure Assign(BaseFilter: IBaseFilter);
    function First: IPin;
    function IndexOf(Item: IPin): Integer;
    function Add(Item: IPin): Integer;
    procedure Insert(Index: Integer; Item: IPin);
    function Last: IPin;
    function Remove(Item: IPin): Integer;
    property Items[Index: Integer]: IPin read GetPin write PutPin; default;
    property PinInfo[Index: Integer]: TPin_Info read GetPinInfo;
  end;

//******************************************************************************
//
//  TFilterSampleGrabber declaration
//  description: Sample Grabber Wrapper Filter
//
//******************************************************************************

  PBufferCB = ^TBufferCB;
  TBufferCB = record
    buffer: PByte;
    BufferLen: Cardinal;
    SampleTime: Double;
  end;

  TFilterSampleGrabber = class;

  // private use
  TSampleGrabberCB = class(TInterfacedObject, ISampleGrabberCB)
    Parent: TFilterSampleGrabber;
    function  SampleCB(SampleTime: Double; pSample: IMediaSample): HResult; stdcall;
    function  BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: longint): HResult; stdcall;
  end;

  TOnBuffer = procedure(sender: TComponent; Buffer: TBufferCB) of object ;

  TFilterSampleGrabber = class(TDSFilterConnector)
  private
    BMPInfo : PBitmapInfo;
    FHandle : THandle;
    FOnBuffer: TOnBuffer;
    SampleGrabberCB: TSampleGrabberCB;
    procedure WndProc(var Msg: TMessage);
  protected
  //  procedure DoEvent(Event, Param1, Param2: longint); dynamic;
  public
    SampleGrabber: ISampleGrabber;
    InPut        : IPin;
    OutPut       : IPin;
    // constructor - destructor
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // functions
    procedure SetBMPCompatible(Source: PAM_Media_Type; default: cardinal);
    function GetBitmap(Bitmap: TBitmap): boolean; overload;
    function GetBitmap(Bitmap: TBitmap; Buffer: TBufferCB): boolean; overload;
  published
    property OnBuffer: TOnBuffer read FOnBuffer write FOnBuffer;
    property Handle  : THandle   read FHandle;
  end;

//******************************************************************************
//
// TFilterFileSourceAsync  declaration
// description: File Source (Async) Wrapper Filter
//
//******************************************************************************

  TFilterFileSourceAsync = class(TDSFilterConnector)
  private
    FFileName: TFileName;
    procedure SetFileName(path: TFileName);
  public
    FileSource : IFileSourceFilter;
    AsyncReader: IAsyncReader;
    OutPut     : IPin;
    MediaType  : TDSMediaType;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure RenderFile;
  published
    property FileName : TFileName read FFileName write SetFileName;
  end;

//******************************************************************************
//
//  TFilterWDMVideoCapture declaration
//  description: WDM Video Capture Wrapper Filter
//
//******************************************************************************

  TFilterWDMVideoCapture = class(TDSFilterConnector)
  private
    FilterList : TList;
    function GetCount: integer;
    function GetFilterName(index: integer): string;
  public
    Pins : TDSPinList;
    ExtDevice          : IAMExtDevice;
    ExtTransport       : IAMExtTransport;
    TimecodeReader     : IAMTimecodeReader;
    CameraControl      : IAMCameraControl;
    VideoProcAmp       : IAMVideoProcAmp;
    AnalogVideoDecoder : IAMAnalogVideoDecoder;
    VideoCompression   : IAMVideoCompression;
    VideoControl       : IAMVideoControl;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function SelectDevice(index: integer): boolean;
    property Count: integer read GetCount;
    property DeviceName[index: integer]: string read GetFilterName;
  published

  end;

//******************************************************************************
//
// TFilterVFWVideoCapture declaration
// description: VFW Video Capture Wrapper Filter
// Attention: You can't enumerate pins if the Filter is not within a FilterGraph
//******************************************************************************

  TFilterVFWVideoCapture = class(TDSFilterConnector)
  private
    FilterList : TList;
    function GetCount: integer;
    function GetFilterName(index: integer): string;
  public
    Pins : TDSPinList;
    CaptureDialogs : IAMVfwCaptureDialogs;
    FilterMiscFlags: IAMFilterMiscFlags;
    OverlayNotify  : IOverlayNotify;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function SelectDevice(index: integer): boolean;
    property Count: integer read GetCount;
    property DeviceName[index: integer]: string read GetFilterName;
  end;

//******************************************************************************
//
// TFilterASFWriter declaration
//
//******************************************************************************

  TFilterASFWriter  = class(TDSFilterConnector)
  private
    FPort: DWORD;
    FMaxUsers: DWORD;
    FProfile: TWMPofiles;
    FFileName: TFileName;
    function GetProfile: TWMPofiles;
    procedure SetProfile(profile: TWMPofiles);
    function GetFileName: TFileName;
    procedure SetFileName(FileName: TFileName);
  protected
    procedure DoOnFilterInGraph(connected: boolean); override;
    procedure Loaded; override;
  public
    FilterMiscFlags      : IAMFilterMiscFlags;
    ConfigAsfWriter      : IConfigAsfWriter;
    FileSinkFilter2      : IFileSinkFilter2;
    MediaSeeking         : IMediaSeeking;
    PersistStream        : IPersistStream;
    ServiceProvider      : IServiceProvider;
    SpecifyPropertyPages : ISpecifyPropertyPages;
    WriterAdvanced2      : IWMWriterAdvanced2;
    WriterNetworkSink    : IWMWriterNetworkSink;
    AudioInput           : IPin;
    VideoInput           : IPin;
    AudioStreamConfig    : IAMStreamConfig;
    VideoStreamConfig    : IAMStreamConfig;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Profile: TWMPofiles read GetProfile write SetProfile;
    property FileName: TFileName read GetFileName write SetFileName;
    property Port: DWORD read FPort write FPort;
    property MaxUsers: DWORD read FMaxUsers write FMaxUsers;
  end;

  procedure Register;

implementation
  const
    IID_IPersistStream   : TGUID = '{00000109-0000-0000-C000-000000000046}';
    IID_IServiceProvider : TGUID = '{6d5140c1-7436-11ce-8034-00aa006009fa}';
//    NETWRITE_ASYNC_EVENT : PCHAR = '{6d12fe9b-d029-4d08-b2eb-92c8cab323c7}';

  type
    PFList = ^TFList;
    TFList = record
      name : string;
      index: cardinal;
    end;

//******************************************************************************
//
// TFilterASFWriter implementation
//
//******************************************************************************

  constructor TFilterASFWriter.Create(AOwner: TComponent);
  var
    hr: HRESULT;
  begin
    inherited Create(AOwner);
    if not (csDesigning in ComponentState) then
      begin
        hr := cocreateinstance(CLSID_WMAsfWriter, nil, CLSCTX_INPROC ,IID_IBASEFilter, BaseFilter);
        assert(hr = S_OK, format('Error $%x creating the File Source Async filter',[hr]));
        if assigned(BaseFilter) then
        begin
          BaseFilter.QueryInterface(IID_IAMFilterMiscFlags    , FilterMiscFlags);
          BaseFilter.QueryInterface(IID_IConfigAsfWriter      , ConfigAsfWriter);
          BaseFilter.QueryInterface(IID_IFileSinkFilter2      , FileSinkFilter2);
          BaseFilter.QueryInterface(IID_IMediaSeeking         , MediaSeeking);
          BaseFilter.QueryInterface(IID_IPersistStream        , PersistStream);
          BaseFilter.QueryInterface(IID_IServiceProvider      , ServiceProvider);
          BaseFilter.QueryInterface(IID_ISpecifyPropertyPages , SpecifyPropertyPages);
          if Filename <> '' then
            FileSinkFilter2.SetFileName(StringToOleStr(Filename),nil);
        end;
    end;
  end;

  destructor TFilterASFWriter.Destroy;
  begin
    if not (csDesigning in ComponentState) then
      begin
        FilterMiscFlags      := nil;
        ConfigAsfWriter      := nil;
        FileSinkFilter2      := nil;
        MediaSeeking         := nil;
        PersistStream        := nil;
        ServiceProvider      := nil;
        SpecifyPropertyPages := nil;
      end;
    inherited destroy;
  end;

  procedure TFilterASFWriter.DoOnFilterInGraph(connected: boolean);
  var PinList: TDSPinList;
  begin
    if connected then
    begin
      if assigned(BaseFilter) then
      begin
        PinList:= TDSPinList.Create(BaseFilter);
        if PinList.Count > 1 then
        begin
          AudioInput := PinList.Items[0];
          VideoInput := PinList.Items[1];
          AudioInput.QueryInterface(IID_IAMStreamConfig, AudioStreamConfig);
          VideoInput.QueryInterface(IID_IAMStreamConfig, VideoStreamConfig);
          if assigned(ServiceProvider) then
            ServiceProvider.QueryService(IID_IWMWriterAdvanced2, IID_IWMWriterAdvanced2, WriterAdvanced2);
          if ((FPort > 0) and (FMaxUsers > 0)) then
          begin
            WMCreateWriterNetworkSink(WriterNetworkSink);
            WriterNetworkSink.SetNetworkProtocol(WMT_PROTOCOL_HTTP);
            WriterNetworkSink.SetMaximumClients(FMaxUsers);
            WriterNetworkSink.Open(FPort);
            WriterAdvanced2.AddSink(WriterNetworkSink);
          end;
        end;
        PinList.Free;
      end;
    end
    else
    begin
      AudioInput := nil;
      VideoInput := nil;
      AudioStreamConfig := nil;
      VideoStreamConfig := nil;
      if assigned(WriterNetworkSink) then WriterAdvanced2.RemoveSink(WriterNetworkSink);
      WriterNetworkSink := nil;
      WriterAdvanced2   := nil;
    end;
    inherited DoOnFilterInGraph(connected);
  end;

  function TFilterASFWriter.GetProfile: TWMPofiles;
  var GUIDProf: TGUID;
  begin
    if  (csLoading in ComponentState)or(csDesigning in ComponentState) then
      result := FProfile
    else
    begin
      ConfigAsfWriter.GetCurrentProfileGuid(GUIDProf);
      result := GuidToProfile(GUIDProf);
    end;
  end;

  procedure TFilterASFWriter.SetProfile(profile: TWMPofiles);
  begin
    if  (csLoading in ComponentState)or(csDesigning in ComponentState) then
      FProfile := profile
    else
      ConfigAsfWriter.ConfigureFilterUsingProfileGuid(ProfileToGUID(profile));
  end;

  function TFilterASFWriter.GetFileName: TFileName;
  var F: PWideChar;
  begin
    if  (csLoading in ComponentState)or(csDesigning in ComponentState) then
      result := FFileName
    else
    begin
      FileSinkFilter2.GetCurFile(F,nil);
      result := F;
    end;
  end;

  procedure TFilterASFWriter.SetFileName(FileName: TFileName);
  begin
    if  (csLoading in ComponentState)or(csDesigning in ComponentState) then
      FFileName := FileName
    else
      FileSinkFilter2.SetFileName(StringToOleStr(FileName),nil);
  end;

  procedure TFilterASFWriter.Loaded;
  begin
    inherited loaded;
    self.SetProfile(FProfile);
    self.SetFileName(FFileName);
  end;

//******************************************************************************
//
//  TFilterVFWVideoCapture implementation
//
//******************************************************************************

  constructor TFilterVFWVideoCapture.Create(AOwner: TComponent);
  var
    EnumMoniker : IEnumMoniker;
    Moniker     : IMoniker;
    PropertyBag : IPropertyBag;
    SysDevEnum  : ICreateDevEnum;
    FriendlyName: OleVariant;
    ACLSID      : OleVariant;
    n           : integer;
    AFList      : PFList;
  begin
    inherited Create(AOwner);
    FilterList := TList.Create;
    Pins := TDSPinList.Create;
    CocreateInstance(CLSID_SystemDeviceEnum, nil, CLSCTX_INPROC, IID_ICreateDevEnum, SysDevEnum);
    SysDevEnum.CreateClassEnumerator(CLSID_VideoInputDeviceCategory, EnumMoniker, 0);
    n := -1;
    while (EnumMoniker.Next(1,Moniker,nil) = S_OK) do
    begin
      n := n + 1;
      moniker.BindToStorage(nil, nil, IID_IPropertyBag, PropertyBag);
      PropertyBag.Read('CLSID', ACLSID, nil);
      if (ACLSID = '{1B544C22-FD0B-11CE-8C63-00AA0044B51E}') then
      // VFW Capture Filter
      begin
        new(AFList);
        PropertyBag.Read('FriendlyName', FriendlyName, nil);
        AFList^.name  := FriendlyName;
        AFList^.index := n;
        FilterList.Add(AFList);
      end;
      PropertyBag := nil;
      Moniker := nil;
    end;
    EnumMoniker.Reset;
    EnumMoniker := nil;
    SysDevEnum  := nil;
  end;

  destructor TFilterVFWVideoCapture.Destroy;
  var i : cardinal;
  begin
    if FilterList.Count > 0 then
    for i := 0 to (FilterList.Count - 1) do dispose(PFList(FilterList.Items[i]));
    FilterList.Clear;
    FilterList.Free;
    Pins.Free;
    BaseFilter := nil;
    inherited Destroy;
  end;

  function TFilterVFWVideoCapture.GetCount: integer;
  begin
    result := FilterList.Count;
  end;

  function TFilterVFWVideoCapture.GetFilterName(index: integer): string;
  begin
    result := '';
    if index < count then result := PFList(FilterList.items[index])^.name;
  end;

  function TFilterVFWVideoCapture.SelectDevice(index: integer): boolean;
  var
    SysDevEnum  : ICreateDevEnum;
    EnumMoniker : IEnumMoniker;
    Moniker     : IMoniker;
  begin
    result := false;
    CocreateInstance(CLSID_SystemDeviceEnum, nil, CLSCTX_INPROC, IID_ICreateDevEnum, SysDevEnum);
    SysDevEnum.CreateClassEnumerator(CLSID_VideoInputDeviceCategory, EnumMoniker, 0);
    EnumMoniker.Skip(PFList(FilterList.items[index])^.index);
    if EnumMoniker.Next(1,moniker,nil) = S_OK then
      if Moniker.BindToObject(nil, nil, IID_IBaseFilter, BaseFilter) = S_OK then
      begin
        result := true;
        Pins.Assign(BaseFilter);
        EnumMoniker.Reset;
        BaseFilter.QueryInterface(IID_IAMVfwCaptureDialogs, CaptureDialogs);
        BaseFilter.QueryInterface(IID_IAMFilterMiscFlags  , FilterMiscFlags);
        BaseFilter.QueryInterface(IID_IOverlayNotify  , OverlayNotify);
      end;
    SysDevEnum  := nil;
    EnumMoniker := nil;
    Moniker     := nil;
  end;

//******************************************************************************
//
//  TFilterWDMVideoCapture implementation
//
//******************************************************************************

  constructor TFilterWDMVideoCapture.Create(AOwner: TComponent);
  var
    EnumMoniker : IEnumMoniker;
    Moniker     : IMoniker;
    PropertyBag : IPropertyBag;
    SysDevEnum  : ICreateDevEnum;
    FriendlyName: OleVariant;
    ACLSID      : OleVariant;
    n           : integer;
    AFList      : PFList;
  begin
    inherited Create(AOwner);
    FilterList := TList.Create;
    Pins := TDSPinList.Create;
    CocreateInstance(CLSID_SystemDeviceEnum, nil, CLSCTX_INPROC, IID_ICreateDevEnum, SysDevEnum);
    SysDevEnum.CreateClassEnumerator(CLSID_VideoInputDeviceCategory, EnumMoniker, 0);
    n := -1;
    while (EnumMoniker.Next(1,Moniker,nil) = S_OK) do
    begin
      n := n + 1;
      moniker.BindToStorage(nil, nil, IID_IPropertyBag, PropertyBag);
      PropertyBag.Read('CLSID', ACLSID, nil);
      if (ACLSID = '{17CCA71B-ECD7-11D0-B908-00A0C9223196}') then
      // WDM Capture Filter
      begin
        new(AFList);
        PropertyBag.Read('FriendlyName', FriendlyName, nil);
        AFList^.name  := FriendlyName;
        AFList^.index := n;
        FilterList.Add(AFList);
      end;
      PropertyBag := nil;
      Moniker := nil;
    end;
    EnumMoniker.Reset;
    EnumMoniker := nil;
    SysDevEnum  := nil;
  end;

  destructor TFilterWDMVideoCapture.Destroy;
  var i : cardinal;
  begin
    if FilterList.Count > 0 then
    for i := 0 to (FilterList.Count - 1) do dispose(PFList(FilterList.Items[i]));
    FilterList.Clear;
    FilterList.Free;
    Pins.Free;
    BaseFilter := nil;
    inherited Destroy;
  end;

  function TFilterWDMVideoCapture.GetCount: integer;
  begin
    result := FilterList.Count;
  end;

  function TFilterWDMVideoCapture.GetFilterName(index: integer): string;
  begin
    result := '';
    if index < count then result := PFList(FilterList.items[index])^.name;
  end;

  function TFilterWDMVideoCapture.SelectDevice(index: integer): boolean;
  var
    SysDevEnum  : ICreateDevEnum;
    EnumMoniker : IEnumMoniker;
    Moniker     : IMoniker;
  begin
    result := false;
    CocreateInstance(CLSID_SystemDeviceEnum, nil, CLSCTX_INPROC, IID_ICreateDevEnum, SysDevEnum);
    SysDevEnum.CreateClassEnumerator(CLSID_VideoInputDeviceCategory, EnumMoniker, 0);
    EnumMoniker.Skip(PFList(FilterList.items[index])^.index);
    if EnumMoniker.Next(1,moniker,nil) = S_OK then
      if Moniker.BindToObject(nil, nil, IID_IBaseFilter, BaseFilter) = S_OK then
      begin
        result := true;
        Pins.Assign(BaseFilter);
        EnumMoniker.Reset;
        BaseFilter.QueryInterface(IID_IAMExtDevice         , ExtDevice);
        BaseFilter.QueryInterface(IID_IAMExtTransport      , ExtTransport);
        BaseFilter.QueryInterface(IID_IAMTimecodeReader    , TimecodeReader);
        BaseFilter.QueryInterface(IID_IAMCameraControl     , CameraControl);
        BaseFilter.QueryInterface(IID_IAMVideoProcAmp      , VideoProcAmp);
        BaseFilter.QueryInterface(IID_IAMAnalogVideoDecoder, AnalogVideoDecoder);
        BaseFilter.QueryInterface(IID_IAMVideoCompression  , VideoCompression);
        BaseFilter.QueryInterface(IID_IAMVideoControl      , VideoControl);
      end;
    SysDevEnum  := nil;
    EnumMoniker := nil;
    Moniker     := nil;
  end;

//******************************************************************************
//
//  TFilterFileSourceAsync implementation
//
//******************************************************************************

  procedure TFilterFileSourceAsync.RenderFile;
  begin
    if Assigned(Graph) then
      if Graph.Connected then
        if FileName <> '' then
          Graph.GraphBuilder.Render(Output);
  end;

  procedure TFilterFileSourceAsync.SetFileName(path: TFileName);
  var
    hr : HRESULT;
    EnumPins: IEnumPins;
    PFile: PWideChar;
  begin
    if csDesigning in ComponentState then
    begin
      FFileName := path;
      exit;
    end;
    if FFileName <> '' then exit;
    BaseFilter.QueryInterface(IID_IFileSourceFilter,FileSource);
    hr := FileSource.Load(StringToOleStr(path),nil);
    if hr = S_OK then
    begin
      BaseFilter.EnumPins(EnumPins);
      EnumPins.Next(1,Output,nil);
      Output.QueryInterface(IID_IAsyncReader,AsyncReader);
      EnumPins := nil;
    end;
    FFileName := path;
    MediaType := TDSMediaType.Create;
    hr := FileSource.GetCurFile(PFile,MediaType.AMMediaType);
    if hr <> S_OK then FFileName := '';
  end;

  constructor TFilterFileSourceAsync.Create(AOwner: TComponent);
  var
    hr : HRESULT;
  begin
    inherited Create(AOwner);
    if not (csDesigning in ComponentState) then
    begin
      hr := cocreateinstance(CLSID_AsyncReader, nil, CLSCTX_INPROC ,IID_IBASEFilter, BaseFilter);
      assert(hr = S_OK, format('Error $%x creating the File Source Async filter',[hr]));
      if FFileName <> '' then SetFileName(FFileName);
    end;
  end;

  destructor  TFilterFileSourceAsync.Destroy;
  begin
    if assigned(MediaType) then MediaType.Free;
    AsyncReader:= nil;
    OutPut     := nil;
    FileSource := nil;
    BaseFilter := nil;
    inherited Destroy;
  end;

//******************************************************************************
//
//  TFilterSampleGrabber implementation
//
//******************************************************************************

  constructor TFilterSampleGrabber.Create(AOwner: TComponent);
  var
    hr : HRESULT;
    EnumPins: IEnumPins;
  begin
    inherited Create(AOwner);
    FHandle := AllocateHWnd(WndProc);
    hr := cocreateinstance(CLSID_SampleGrabber, nil, CLSCTX_INPROC ,IID_IBASEFilter, BaseFilter);
    assert(hr = S_OK, format('Error $%x creating the sample grabber filter',[hr]));
    BaseFilter.QueryInterface(IID_ISampleGrabber,SampleGrabber);
    BaseFilter.EnumPins(EnumPins);
    EnumPins.Next(1,InPut,nil);
    EnumPins.Next(1,OutPut,nil);
    EnumPins := nil;
    SampleGrabber.SetBufferSamples(true);
    SampleGrabberCB := TSampleGrabberCB.Create;
    SampleGrabberCB.Parent := self;
    SampleGrabber.SetCallback(SampleGrabberCB,1);
    new(BMPInfo);
  end;

  destructor TFilterSampleGrabber.Destroy;
  begin
    dispose(BMPInfo);
    BaseFilter:= nil;
    SampleGrabber:= nil;
    DeallocateHWnd(FHandle);
    inherited destroy;
  end;

  procedure TFilterSampleGrabber.WndProc(var Msg: TMessage);
  begin
    with Msg do
        if Msg = WM_CAPTURE_BITMAP then
          try
            if assigned(FOnBuffer) then
              FOnBuffer(self, PBufferCB(WParam)^);
            freemem(PBufferCB(WParam)^.buffer, PBufferCB(WParam)^.BufferLen);
            dispose(PBufferCB(WParam));
          except
            Application.HandleException(Self);
          end
        else
          Result := DefWindowProc(FHandle, Msg, wParam, lParam);
  end;

  // Set the Sample Grabber Media Type Compatible with RGB Bitmap format
  // parameter: Source Media Type
  procedure TFilterSampleGrabber.SetBMPCompatible(Source: PAM_Media_Type; default: cardinal);
  var
    PMT     : PAM_Media_Type;
    SubType : TGUID;
    BitCount: LongWord;
  begin
    BitCount := default;
      new(PMT);
      ZeroMemory(PMT, sizeof(TAM_MEDIA_TYPE)); // very important !!
      PMT^.majortype := MEDIATYPE_Video;
      PMT^.formattype := FORMAT_VideoInfo;
      if Source = nil then
      begin
        case default of
          0      : PMT^.subtype := MEDIASUBTYPE_RGB24;
          1      : PMT^.subtype := MEDIASUBTYPE_RGB1;
          2 ..4  : PMT^.subtype := MEDIASUBTYPE_RGB4;
          5 ..8  : PMT^.subtype := MEDIASUBTYPE_RGB8;
          9 ..16 : PMT^.subtype := MEDIASUBTYPE_RGB555;
          17..24 : PMT^.subtype := MEDIASUBTYPE_RGB24;
          25..32 : PMT^.subtype := MEDIASUBTYPE_RGB32
        else
          PMT^.subtype := MEDIASUBTYPE_RGB32;
        end;
        SampleGrabber.SetMediaType(PMT^);
        dispose(PMT);
        exit;
      end;

      SubType := Source.subtype;
      if (IsEqualGUID(SubType, MEDIASUBTYPE_RGB1)   or
          IsEqualGUID(SubType, MEDIASUBTYPE_RGB4)   or
          IsEqualGUID(SubType, MEDIASUBTYPE_RGB8)   or
          IsEqualGUID(SubType, MEDIASUBTYPE_RGB555) or
          IsEqualGUID(SubType, MEDIASUBTYPE_RGB24)  or
          IsEqualGUID(SubType, MEDIASUBTYPE_RGB32)) then
            PMT^.subtype := SubType // no change
      else
      begin
        // get bitcount
        if assigned(Source.pbFormat) then
        if IsEqualGUID(Source.formattype, FORMAT_VideoInfo) then
          BitCount := PVideoInfoHeader(Source.pbFormat)^.bmiHeader.biBitCount    else
        if IsEqualGUID(Source.formattype, FORMAT_VideoInfo2) then
          BitCount := PVideoInfoHeader2(Source.pbFormat)^.bmiHeader.biBitCount   else
        if IsEqualGUID(Source.formattype, FORMAT_MPEGVideo) then
          BitCount := PMPEG1VideoInfo(Source.pbFormat)^.hdr.bmiHeader.biBitCount else
        if IsEqualGUID(Source.formattype, FORMAT_MPEG2Video) then
          BitCount := PMPEG2VideoInfo(Source.pbFormat)^.hdr.bmiHeader.biBitCount;
        case BitCount of
          0      : PMT^.subtype := MEDIASUBTYPE_RGB24;
          1      : PMT^.subtype := MEDIASUBTYPE_RGB1;
          2 ..4  : PMT^.subtype := MEDIASUBTYPE_RGB4;
          5 ..8  : PMT^.subtype := MEDIASUBTYPE_RGB8;
          9 ..16 : PMT^.subtype := MEDIASUBTYPE_RGB555;
          17..24 : PMT^.subtype := MEDIASUBTYPE_RGB24;
          25..32 : PMT^.subtype := MEDIASUBTYPE_RGB32
        else
          PMT^.subtype := MEDIASUBTYPE_RGB32;
        end;
      end;
      SampleGrabber.SetMediaType(PMT^);
      dispose(PMT);
  end;

  function TFilterSampleGrabber.GetBitmap(Bitmap: TBitmap): boolean;
  var
    hr         : HRESULT;
    BMIHeader  : TBitmapInfoHeader;
    pBufferSize: longint;
    MediaType  : TAM_Media_Type;
    buffer     : pointer;
  begin
    result := false;
    if not assigned(Bitmap) then exit;
    hr := SampleGrabber.GetConnectedMediaType(MediaType);
    if (hr <> S_OK) then exit;
    if IsEqualGUID(MediaType.majortype, MEDIATYPE_Video) then
    begin
      case mediatype.formattype.D1 of
        $05589F80: BMIHeader := PVideoInfoHeader(mediatype.pbFormat)^.bmiHeader;
        $F72A76A0: BMIHeader := PVideoInfoHeader2(mediatype.pbFormat)^.bmiHeader;
      end;
      ZeroMemory(BMPInfo, sizeof(TBitmapInfo));
      CopyMemory(@BMPInfo.bmiHeader, @bmiHeader, sizeof(TBITMAPINFOHEADER));
      bitmap.Handle := CreateDIBSection(0, BMPInfo^, DIB_RGB_COLORS, buffer, 0, 0);
      hr := SampleGrabber.GetCurrentBuffer(pBufferSize, buffer);
      if (hr <> S_OK) then exit;
      result := true;
    end;
    FreeMediaType(@MediaType);
  end;

  function TFilterSampleGrabber.GetBitmap(Bitmap: TBitmap; Buffer: TBufferCB): boolean;
  var
    hr         : HRESULT;
    BMIHeader  : TBitmapInfoHeader;
    MediaType  : TAM_Media_Type;
    buf        : pointer;
  begin
    result := false;
    if ((Buffer.buffer = nil) or (buffer.BufferLen = 0)) then exit;
    if not assigned(Bitmap) then exit;
    hr := SampleGrabber.GetConnectedMediaType(MediaType);
    if (hr <> S_OK) then exit;
    if IsEqualGUID(MediaType.majortype, MEDIATYPE_Video) then
    begin
      case mediatype.formattype.D1 of
        $05589F80: BMIHeader := PVideoInfoHeader(mediatype.pbFormat)^.bmiHeader;
        $F72A76A0: BMIHeader := PVideoInfoHeader2(mediatype.pbFormat)^.bmiHeader;
      end;
      FillChar(BMPInfo^, 44, 0);
      Move(bmiHeader, BMPInfo.bmiHeader, 40);
      bitmap.Handle := CreateDIBSection(0, BMPInfo^, DIB_RGB_COLORS, buf, 0, 0);
      if buf = nil then exit;
      Move(buffer.buffer^, buf^, buffer.BufferLen);
      result := true;
    end;
    FreeMediaType(@MediaType);
  end;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  function TSampleGrabberCB.SampleCB(SampleTime: Double; pSample: IMediaSample): HResult; stdcall;
  begin
    result := 0;
  end;

  function TSampleGrabberCB.BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: longint): HResult; stdcall;
  var buf: PBufferCB;
  begin
    if assigned(parent.FOnBuffer) then
    begin
      new(buf);
      getmem(buf^.buffer,BufferLen);
      buf^.BufferLen  := BufferLen;
      buf^.SampleTime := SampleTime;
      copymemory(buf^.buffer,pBuffer,BufferLen);
      SendMessage(parent.FHandle, WM_CAPTURE_BITMAP, cardinal(buf), 0);
    end;
    result := 0;
  end;

//******************************************************************************
//
//  TDSPinList implementation
//
//******************************************************************************

  constructor TDSPinList.Create(BaseFilter: IBaseFilter);
  begin
    inherited Create;
    Filter := BaseFilter;
    Update;
  end;

  destructor TDSPinList.Destroy;
  begin
    inherited Destroy;
  end;

  procedure TDSPinList.Update;
  var EnumPins: IEnumPins;
      Pin: IPin;
  begin
    if assigned(Filter) then
    Filter.EnumPins(EnumPins);
    while (EnumPins.Next(1, pin, nil) = S_OK) do add(Pin);
    EnumPins := nil;
  end;

  procedure TDSPinList.Assign(BaseFilter: IBaseFilter);
  begin
    Clear;
    Filter := BaseFilter;
    Update;
  end;

  function TDSPinList.GetPin(Index: Integer): IPin;
  begin
    result := get(index) as IPin;
  end;

  procedure TDSPinList.PutPin(Index: Integer; Item: IPin);
  begin
    put(index,Item);
  end;

  function TDSPinList.First: IPin;
  begin
    result := GetPin(0);
  end;

  function TDSPinList.IndexOf(Item: IPin): Integer;
  begin
     result := inherited IndexOf(Item);
  end;

  function TDSPinList.Add(Item: IPin): Integer;
  begin
    result := inherited Add(Item);
  end;

  procedure TDSPinList.Insert(Index: Integer; Item: IPin);
  begin
    inherited Insert(index,item);
  end;

  function TDSPinList.Last: IPin;
  begin
    result := inherited Last as IPin;
  end;

  function TDSPinList.Remove(Item: IPin): Integer;
  begin
    result := inherited Remove(Item);
  end;

  function TDSPinList.GetPinInfo(index: integer): TPin_Info;
  begin
    if assigned(Items[index]) then Items[index].QueryPinInfo(result);
  end;

////////////////////////////////////////////////////////////////////////////////

  function GetBitmapSize(const pHeader: TBITMAPINFOHEADER): DWORD;
    function WIDTHBYTES(bits: DWORD): DWORD;
      begin result := DWORD((bits+31) and (not 31)) div 8; end;
    function DIBWIDTHBYTES(bi: TBITMAPINFOHEADER): DWORD;
      begin result := DWORD(WIDTHBYTES(DWORD(bi.biWidth) * DWORD(bi.biBitCount))); end;
    function _DIBSIZE(bi: TBITMAPINFOHEADER): DWORD;
      begin result := DIBWIDTHBYTES(bi) * DWORD(bi.biHeight); end;
  begin
    if (pHeader.biHeight < 0) then result := -1 * _DIBSIZE(pHeader)
    else result := _DIBSIZE(pHeader);
  end;

  function PALETTISED(pbmi: TVIDEOINFO): boolean;
  begin
    result := pbmi.bmiHeader.biBitCount <= iPALETTE;
  end;

  function PALETTE_ENTRIES(pbmi: TVIDEOINFO): DWORD;
  begin
    result := DWORD(1) shl pbmi.bmiHeader.biBitCount;
  end;

//******************************************************************************

// REGISTER
  procedure Register;
  begin
    RegisterComponents('Direct Show', [TDSFilterGraph2, TDSDVDGraph, TDSCaptureGraph,
      TDSVideoWindow {$IFNDEF VER120},  TDSTrackBar {$ENDIF}, TFilterSampleGrabber,
      TFilterFileSourceAsync, TFilter, TFilterASFWriter]);
  end;

end.
